// Copyright (C) 2010  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_MANIFOLD_REGULARIzATION_HEADER
#define DLIB_MANIFOLD_REGULARIzATION_HEADER

#include "graph_utils/edge_list_graphs.h"
#include "manifold_regularization/linear_manifold_regularizer.h"
#include "graph_utils/function_objects.h"

#endif // DLIB_MANIFOLD_REGULARIzATION_HEADER



